# -*- coding: utf-8 -*-

# URLhausCheck.py - version 1.10
# developed my Lars Blomgaard in collaboraration with chat GPT and Genemi
# date: 09-04-2025

import os
import csv
import inspect

from java.lang import System
from java.util.logging import Level
from java.io import File, BufferedReader, InputStreamReader, BufferedInputStream, FileInputStream
from java.io import FileOutputStream, OutputStreamWriter, PrintWriter
from java.net import URL as JavaURL
from java.sql import DriverManager
from java.util.zip import ZipInputStream

from org.sleuthkit.datamodel import BlackboardArtifact, BlackboardAttribute, Content
from org.sleuthkit.autopsy.ingest import IngestModuleFactoryAdapter, DataSourceIngestModule, IngestModule
from org.sleuthkit.autopsy.ingest.IngestModule import ProcessResult
from org.sleuthkit.autopsy.ingest import IngestMessage, IngestServices, ModuleDataEvent
from org.sleuthkit.autopsy.coreutils import Logger
from org.sleuthkit.autopsy.casemodule import Case


class URLhausCheckModuleFactory(IngestModuleFactoryAdapter):
    moduleName = "URLhaus Lookup Module"

    def getModuleDisplayName(self):
        return self.moduleName

    def getModuleDescription(self):
        return "Checks browser URL artifacts against the URLhaus threat database."

    def getModuleVersionNumber(self):
        # <<< UPDATED: Incremented version number
        return "1.10"

    def isDataSourceIngestModuleFactory(self):
        return True

    def createDataSourceIngestModule(self, settings):
        return URLhausCheckModule(self.moduleName)

class URLhausCheckModule(DataSourceIngestModule):
    def __init__(self, moduleName):
         self._logger = Logger.getLogger(moduleName)
         self.moduleName = moduleName

    def log(self, level, msg):
        self._logger.logp(level, self.__class__.__name__, inspect.stack()[1][3], msg)

    def extract_domain(self, url_string):
        domain = None
        if not url_string: return None
        try:
            if not url_string.startswith(('http://', 'https://', 'ftp://')):
                 processed_url_str = 'http://' + url_string
            else:
                 processed_url_str = url_string
            if processed_url_str.startswith('//'):
                 processed_url_str = 'http:' + processed_url_str
            java_url_obj = JavaURL(processed_url_str)
            host = java_url_obj.getHost()
            if host: domain = host.lower()
        except Exception as e:
            self.log(Level.WARNING, "Error parsing URL '{}' for domain: {}".format(url_string, str(e)))
        return domain

    def startUp(self, context):
        self.context = context
        self.malicious_domains = set()
        self.tsk_url_type = None
        self.tsk_set_name_type = None
        self.tsk_associated_artifact_type = None
        self.tsk_domain_type = None
        self.tsk_comment_type = None
        self.log(Level.INFO, "Starting up URLhaus module.")
        module_dir = Case.getCurrentCase().getModuleDirectory()
        self.urlhaus_csv = os.path.join(module_dir, "urlhaus.csv")
        if not os.path.exists(module_dir):
             self.log(Level.SEVERE, "Module directory not found: {}".format(module_dir))
             raise Exception("Module directory not found.")
        if not os.path.exists(self.urlhaus_csv):
            self.log(Level.SEVERE, "URLhaus CSV file not found at: {}".format(self.urlhaus_csv))
            raise Exception("URLhaus CSV file not found.")
        try:
            with open(self.urlhaus_csv, "r") as f:
                reader = csv.reader(f, delimiter=',', quotechar='"')
                line_num = 0
                for row in reader:
                    line_num += 1
                    if not row or row[0].startswith("#"): continue
                    if len(row) >= 3:
                        url_from_csv = row[2]
                        domain = self.extract_domain(url_from_csv)
                        if domain: self.malicious_domains.add(domain)
                    else:
                         self.log(Level.WARNING, "Skipping row {} in URLhaus CSV: insufficient columns.".format(line_num))
        except IOError as e:
             self.log(Level.SEVERE, "IOError reading URLhaus CSV {}: {}".format(self.urlhaus_csv, str(e)))
             raise Exception("Failed to read URLhaus CSV file.")
        except Exception as e:
            self.log(Level.SEVERE, "Error processing URLhaus CSV {}: {}".format(self.urlhaus_csv, str(e)))
            raise Exception("Failed to process URLhaus CSV file.")
        if not self.malicious_domains:
             self.log(Level.WARNING, "No malicious domains loaded from URLhaus CSV.")
        else:
             self.log(Level.INFO, "Loaded {} unique domains/hosts from URLhaus.".format(len(self.malicious_domains)))
        skCase = Case.getCurrentCase().getSleuthkitCase()
        try:
             self.tsk_url_type = skCase.getAttributeType("TSK_URL")
             self.tsk_set_name_type = skCase.getAttributeType("TSK_SET_NAME")
             self.tsk_associated_artifact_type = skCase.getAttributeType("TSK_ASSOCIATED_ARTIFACT")
             self.tsk_domain_type = skCase.getAttributeType("TSK_DOMAIN")
             self.tsk_comment_type = skCase.getAttributeType("TSK_COMMENT")
             missing_critical = []
             if not self.tsk_url_type: missing_critical.append("TSK_URL")
             if not self.tsk_set_name_type: missing_critical.append("TSK_SET_NAME")
             if not self.tsk_associated_artifact_type: missing_critical.append("TSK_ASSOCIATED_ARTIFACT")
             if missing_critical:
                  msg = "Could not find critical standard attribute types: {}. Module cannot function.".format(missing_critical)
                  self.log(Level.SEVERE, msg)
                  raise Exception(msg)
             if not self.tsk_domain_type: self.log(Level.WARNING, "Standard attribute type TSK_DOMAIN not found.")
             if not self.tsk_comment_type: self.log(Level.WARNING, "Standard attribute type TSK_COMMENT not found.")
        except Exception as e:
             self.log(Level.SEVERE, "Error during attribute type fetching in startup: {}".format(str(e)))
             raise Exception("Failed to get standard attribute types during startup.")
        self.log(Level.INFO, "URLhaus module startup complete.")


    def process(self, dataSource, progressBar):
        if not self.malicious_domains:
             self.log(Level.WARNING, "Skipping processing: no malicious domains loaded.")
             return ProcessResult.OK

        url_hits = 0
        case = Case.getCurrentCase()
        skCase = case.getSleuthkitCase()
        tsk_url_type = self.tsk_url_type
        tsk_set_name_type = self.tsk_set_name_type
        tsk_associated_artifact_type = self.tsk_associated_artifact_type
        tsk_domain_type = self.tsk_domain_type
        tsk_comment_type = self.tsk_comment_type

        if not tsk_url_type or not tsk_set_name_type or not tsk_associated_artifact_type:
             self.log(Level.SEVERE, "Critical attribute types missing at process start. Aborting.")
             return ProcessResult.ERROR

        self.log(Level.INFO, "Scanning TSK_WEB_HISTORY artifacts for malicious domains/hosts...")
        artifacts = skCase.getBlackboardArtifacts(BlackboardArtifact.ARTIFACT_TYPE.TSK_WEB_HISTORY)
        match_list = []
        progressBar.switchToIndeterminate()
        processed_count = 0

        for art in artifacts:
             if self.context.isJobCancelled():
                self.log(Level.INFO, "URLhaus scan cancelled by user.")
                return ProcessResult.OK
             processed_count += 1

             url_attr = art.getAttribute(tsk_url_type)
             if url_attr:
                 url = url_attr.getValueString()
                 if not url: continue
                 domain = self.extract_domain(url)

                 if domain and domain in self.malicious_domains:
                     self.log(Level.INFO, "MATCH: Domain/Host '{}' from URL '{}' found.".format(domain, url))
                     match_list.append((url, domain))
                     try:
                         src_content = art.getParent()
                         if src_content is None:
                              self.log(Level.WARNING, "Could not get parent Content for artifact ID {}. Skipping hit creation.".format(art.getArtifactID()))
                              continue

                         hit_art = src_content.newArtifact(BlackboardArtifact.ARTIFACT_TYPE.TSK_INTERESTING_ARTIFACT_HIT)
                         hit_art.addAttribute(BlackboardAttribute(tsk_set_name_type, self.moduleName, "URLhaus Match"))
                         hit_art.addAttribute(BlackboardAttribute(tsk_associated_artifact_type, self.moduleName, art.getArtifactID()))
                         hit_art.addAttribute(BlackboardAttribute(tsk_url_type, self.moduleName, url))
                         if tsk_domain_type:
                            hit_art.addAttribute(BlackboardAttribute(tsk_domain_type, self.moduleName, domain))
                         if tsk_comment_type:
                            hit_art.addAttribute(BlackboardAttribute(tsk_comment_type, self.moduleName, "Matched domain/host from URLhaus list"))

                         url_hits += 1
                         IngestServices.getInstance().fireModuleDataEvent(
                              ModuleDataEvent(self.moduleName, BlackboardArtifact.ARTIFACT_TYPE.TSK_INTERESTING_ARTIFACT_HIT, [hit_art]))
                     except Exception as e:
                          self.log(Level.SEVERE, "Failed to create TSK_INTERESTING_ARTIFACT_HIT for URL {}: {}".format(url, str(e)))


        self.log(Level.INFO, "Processed {} TSK_WEB_HISTORY artifacts.".format(processed_count))

        # Write matches to output CSV file
        if match_list:
            # <<< MODIFIED: Get output directory using manual path construction >>>
            try:
                 case_dir = Case.getCurrentCase().getCaseDirectory()
                 # Sanitize module name slightly for use as directory name
                 safe_module_name = self.moduleName.replace(" ", "_").replace(":", "").replace("\\", "").replace("/", "")
                 # Construct path: CaseDir/ModuleOutput/SafeModuleName/
                 output_dir = os.path.join(case_dir, "ModuleOutput", safe_module_name)
                 self.log(Level.INFO, "Constructed module output directory path: {}".format(output_dir))
            except Exception as e:
                 self.log(Level.SEVERE, "Failed to construct module output directory path: {}".format(str(e)))
                 output_dir = None

            if not output_dir:
                 self.log(Level.WARNING, "Output directory path could not be determined. Skipping CSV report generation.")
            else:
                # Ensure the directory exists (create if missing)
                if not os.path.exists(output_dir):
                    try:
                        # Check first, then make for broader compatibility
                        os.makedirs(output_dir)
                        self.log(Level.INFO, "Created module output directory: {}".format(output_dir))
                    except OSError as e:
                         # Handle potential race condition or permission issue
                         self.log(Level.SEVERE, "Failed to create module output directory {}: {}".format(output_dir, str(e)))
                         output_dir = None # Don't proceed if creation failed

                if output_dir: # Check again if directory exists now
                    output_path = os.path.join(output_dir, "urlhaus_output.csv")
                    self.log(Level.INFO, "Writing {} matches to {}".format(len(match_list), output_path))
                    try:
                        out_file = File(output_path)
                        out_stream = FileOutputStream(out_file)
                        writer = PrintWriter(OutputStreamWriter(out_stream, "UTF-8"))
                        writer.println('"URL","Domain/Host"')
                        for url, domain in match_list:
                            quoted_url = '"{}"'.format(url.replace('"', '""'))
                            quoted_domain = '"{}"'.format(domain.replace('"', '""'))
                            writer.println("{},{}".format(quoted_url, quoted_domain))
                        writer.close()
                        self.log(Level.INFO, "Successfully wrote URLhaus output CSV.")
                        IngestServices.getInstance().postMessage(IngestMessage.createMessage(
                             IngestMessage.MessageType.INFO, self.moduleName,
                             "URLhaus Matches Report created: {}".format(output_path), output_path))
                    except Exception as e:
                        self.log(Level.SEVERE, "Failed to write URLhaus output CSV to {}: {}".format(output_path, str(e)))
        else:
             self.log(Level.INFO, "No matches found, skipping output CSV file generation.")

        summary_msg = "URLhaus scan complete. Found {} matches.".format(url_hits)
        IngestServices.getInstance().postMessage(
             IngestMessage.createMessage(IngestMessage.MessageType.INFO, self.moduleName, summary_msg)
        )
        return ProcessResult.OK

    def shutDown(self):
        self.log(Level.INFO, "Shutting down URLhaus module.")